@@ -3,6 +3,7 @@ from __future__ import print_function
 
 import argparse
 import datetime
+import filecmp
 import hashlib
 import os
 import sqlite3
@@ -26,6 +27,25 @@ class Row(sqlite3.Row):
         return '\n'.join(['{}: {}'.format(key, self[key]) for key in self.keys()])
 
 
+class RowList(list):
+
+    def _repr_html_(self):
+        if len(self) == 0:
+            return
+        out = '<table>\n<thead><tr>\n'
+        keys = self[0].keys()
+        for k in keys:
+            out += '<th>{}</th>\n'.format(k)
+        out += '</tr>\n</thead>\n<tbody>\n'
+        for r in self:
+            out += '<tr>\n'
+            for k in keys:
+                out += '<th>{}</th>\n'.format(r[k])
+            out += '</tr>\n'
+        out += '</tbody>\n</table>'
+        return out
+
+
 def _key_val_list(d, separate_nulls=False):
     keys = []
     vals = []
@@ -70,12 +90,23 @@ def _hash_metadata(metadata):
     return h.hexdigest()
 
 
-def add(metadata, db="files.db", wd='.', filename=None, timeout=10, ext='', prefix='', suffix=''):
+def add(metadata, db="files.db", wd='.', filename=None, timeout=10, ext='', prefix='', suffix='', copy_mode=False):
     if filename and (ext or prefix or suffix):
         raise ValueError('ext, prefix, and suffix cannot be specified if filename is specified')
-    for reserved_key in RESERVED_KEYS:
-        if reserved_key in metadata.keys():
-            raise ValueError("filename is reserved")
+    if copy_mode:
+        if 'filename' not in metadata:
+            raise ValueError('filename must be in metadata in copy_mode')
+        if filename is not None:
+            raise ValueError('filename kwarg argument must be None in copy_mode')
+        filename = metadata.pop('filename')
+        if 'time' not in metadata:
+            raise ValueError('time must be in metadata in copy_mode')
+        currtime = metadata.pop('time')
+    else:
+        currtime = datetime.datetime.now()
+        for reserved_key in RESERVED_KEYS:
+            if reserved_key in metadata.keys():
+                raise ValueError("filename is reserved")
     conn = _get_conn(db, wd, timeout=timeout)
     with conn:
         desc = conn.execute("select * from filelist").description
@@ -91,7 +122,7 @@ def add(metadata, db="files.db", wd='.', filename=None, timeout=10, ext='', pref
         keys, vals = _key_val_list(metadata)
         if filename is None:
             filename = '{}{}{}{}'.format(prefix, _hash_metadata(metadata), suffix, ext)
-        conn.execute("insert into filelist (filename, time, " + ', '.join(keys) + ") values (" + ', '.join(["?"] * (len(vals) + 2)) + ")", [filename, datetime.datetime.now()] + vals)
+        conn.execute("insert into filelist (filename, time, " + ', '.join(keys) + ") values (" + ', '.join(["?"] * (len(vals) + 2)) + ")", [filename, currtime] + vals)
     return filename
 
 
@@ -121,7 +152,59 @@ def search(metadata, db="files.db", wd='.', timeout=10, verbose=False, keys_to_p
             rows = conn.execute("select * from filelist").fetchall()
     if verbose:
         _print_rows(rows, keys=keys_to_print)
-    return rows
+    return RowList(rows)
+
+
+def _cmprows(r1, r2):
+    r1 = dict(r1)
+    r2 = dict(r2)
+    for k in (set(r1.keys()) | set(r2.keys())):
+        if r1.get(k) != r2.get(k):
+            return False
+    return True
+
+
+def merge(indb, outdb, wd='.'):
+    rowsin = search({}, db=indb, wd=wd)
+    rowsout = search({}, db=outdb, wd=wd)
+    rowsoutdict = {r['filename']: r for r in rowsout}
+    fnamesout = {r['filename'] for r in rowsout}
+    rowsindict = {r['filename']: r for r in rowsin}
+
+    for fname, row in rowsindict.items():
+        if fname in fnamesout:
+            if not _cmprows(row, rowsoutdict[fname]):
+                raise RuntimeError('{} detected in output database, but with different rows'.format(fname))
+        else:
+            add(dict(row), db=outdb, wd=wd, copy_mode=True)
+
+
+def copy(filename, outdir, db="files.db", wd='.', outdb='files.db', copytype='hardlink'):
+    rowin = search({'filename': filename}, db=db, wd=wd)
+    assert len(rowin) == 1
+    rowin = rowin[0]
+    try:
+        row = search({'filename': filename}, db=outdb, wd=outdir)
+    except FileNotFoundError:
+        add(dict(rowin), db=outdb, wd=outdir, copy_mode=True)
+    else:
+        if len(row) > 1:
+            raise RuntimeError('multiple entries found. this should be impossible')
+        elif len(row) == 1:
+            if not _cmprows(row[0], rowin):
+                raise RuntimeError('filename already appears in output database, but with different parameters')
+        else:
+            add(dict(rowin), db=outdb, wd=outdir, copy_mode=True)
+    outfull = os.path.join(outdir, filename)
+    infull = os.path.join(wd, filename)
+    if os.path.exists(outfull):
+        if not filecmp.cmp(outfull, infull, shallow=False):
+            raise RuntimeError('File already copied, but results not identical')
+    else:
+        if copytype == 'hardlink':
+            os.link(infull, outfull)
+        else:
+            raise ValueError('unsupported copytype')
 
 
 def delete(metadata, db='files.db', wd='.', timeout=10, dryrun=False, delimiter='\t', comparison_operators=None):
@@ -141,7 +224,7 @@ def delete(metadata, db='files.db', wd='.', timeout=10, dryrun=False, delimiter=
                     # and it should be called directly (not in a script) so the user can
                     # keep track of potential issues.
                     conn.execute('delete from filelist where filename=?', (r['filename'],))
-    return rows
+    return RowList(rows)
 
 
 def _parse_metadata(metadatalist):
@@ -206,6 +289,10 @@ def main():
     parser_delete.add_argument('metadata', nargs='*', help='List of keys and values', metavar="KEY=VALUE")
     parser_delete.set_defaults(subcommand='delete')
 
+    parser_merge = subparsers.add_parser('merge', help='Merge input database into --db. (input remains unchanged)')
+    parser_merge.add_argument('input', type=str, help='Input database.')
+    parser_merge.set_defaults(subcommand='merge')
+
     parser_test = subparsers.add_parser('test', help='Run tests')
     parser_test.set_defaults(subcommand='test')
 
@@ -234,6 +321,9 @@ def main():
         _print_rows(rows, delimiter=args.delimiter,
                     keys=None if args.output_columns is None else args.output_columns.split(','))
 
+    elif args.subcommand == 'merge':
+        merge(args.input, args.db, wd=args.wd)
+
     elif args.subcommand == 'test':
         import pytest
         pytest.main([os.path.split(__file__)[0]])
