@@ -1,4 +1,6 @@
 from collections import OrderedDict
+import datetime
+import filecmp
 import pytest
 import os
 import sqlite3
@@ -345,3 +347,142 @@ def test_filename_kwargs(tmpdir):
     out = filesdb.add(dict(field1='test'), wd=str(tmpdir), suffix='hi', prefix='there', ext='.wasd')
     assert out[:5] == 'there'
     assert out[-7:] == 'hi.wasd'
+
+
+def test_repr_html(tmpdir):
+    db = "files.db"
+    filesdb.add(dict(field1="one", field2=2, field3=3.0, field4=True, field5=None), db=db, wd=str(tmpdir))
+    out = filesdb.search({}, wd=str(tmpdir))
+    out._repr_html_()
+    out = filesdb.search(dict(field1="two"), wd=str(tmpdir))
+    out._repr_html_()
+
+
+def test_copy(tmpdir):
+    indir = os.path.join(str(tmpdir), 'indir')
+    os.mkdir(indir)
+    outdir = os.path.join(str(tmpdir), 'outdir')
+    os.mkdir(outdir)
+    fname = filesdb.add(dict(field1="one"), wd=indir)
+    infname = os.path.join(indir, fname)
+    outfname = os.path.join(outdir, fname)
+
+    with open(infname, 'w') as f:
+        f.write('test')
+
+    filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+
+    assert filecmp.cmp(infname, outfname)
+    assert os.stat(infname, follow_symlinks=False).st_ino == os.stat(outfname, follow_symlinks=False).st_ino
+
+    inrows = filesdb.search({}, wd=indir)
+    outrows = filesdb.search({}, wd=outdir)
+    assert len(inrows) == 1
+    assert len(outrows) == 1
+    assert inrows[0] == outrows[0]
+
+    filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+
+    fname = filesdb.add(dict(field1="two"), wd=indir)
+    infname = os.path.join(indir, fname)
+    outfname = os.path.join(outdir, fname)
+    with open(infname, 'w') as f:
+        f.write('test2')
+
+    filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+    assert filecmp.cmp(infname, outfname)
+    assert os.stat(infname, follow_symlinks=False).st_ino == os.stat(outfname, follow_symlinks=False).st_ino
+
+    fname = filesdb.add(dict(field1="three"), wd=indir)
+    filesdb.add(dict(field1="four"), wd=outdir, filename=fname)
+
+    with pytest.raises(RuntimeError):
+        filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+
+    fname = filesdb.add(dict(field1="five"), wd=indir)
+    infname = os.path.join(indir, fname)
+    outfname = os.path.join(outdir, fname)
+    with open(infname, 'w') as f:
+        f.write('test')
+    with open(outfname, 'w') as f:
+        f.write('test2')
+
+    with pytest.raises(RuntimeError):
+        filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+
+
+def test_copy_newcol(tmpdir):
+    indir = os.path.join(str(tmpdir), 'indir')
+    os.mkdir(indir)
+    outdir = os.path.join(str(tmpdir), 'outdir')
+    os.mkdir(outdir)
+    fname = filesdb.add(dict(field1="one"), wd=indir)
+    infname = os.path.join(indir, fname)
+
+    with open(infname, 'w') as f:
+        f.write('test')
+
+    filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+    filesdb.add(dict(field1="two", field2="three"), wd=indir)
+    filesdb.copy(fname, outdir, wd=indir, copytype='hardlink')
+
+
+def test_cmprows():
+    r1 = dict(hi=2, there=3)
+    r2 = dict(hi=2, there=3)
+    assert filesdb._cmprows(r1, r2)
+    r1 = dict(hi=2)
+    r2 = dict(hi=2, there=None)
+    assert filesdb._cmprows(r1, r2)
+    r1 = dict(hi=2, there=1)
+    r2 = dict(hi=2, there=None)
+    assert not filesdb._cmprows(r1, r2)
+    r1 = dict(hi=2, there=1)
+    r2 = dict(hi=2)
+    assert not filesdb._cmprows(r1, r2)
+
+
+def test_add_fail(tmpdir):
+    with pytest.raises(ValueError):
+        filesdb.add({'filename': 'test'}, wd=str(tmpdir), copy_mode=True)
+    with pytest.raises(ValueError):
+        filesdb.add({'time': datetime.datetime.now()}, wd=str(tmpdir), copy_mode=True)
+    with pytest.raises(ValueError):
+        filesdb.add({'filename': 'test', 'time': datetime.datetime.now()}, filename='test', wd=str(tmpdir), copy_mode=True)
+
+
+def test_merge(tmpdir):
+    in1 = 'in1.db'
+    out = 'out.db'
+    filesdb.add(dict(hi=2, there=3), db=in1, wd=str(tmpdir), filename='1')
+    filesdb.add(dict(hi=3, there=3), db=in1, wd=str(tmpdir), filename='2')
+    rowitem2 = [r for r in filesdb.search({}, db=in1, wd=str(tmpdir)) if r['filename'] == '2'][0]
+
+    filesdb.add(dict(hi=5), wd=str(tmpdir), db=out, filename='5')
+    filesdb.add(dict(rowitem2), db=out, wd=str(tmpdir), copy_mode=True)
+    filesdb.merge(in1, out, wd=str(tmpdir))
+    r1 = filesdb.search({}, db=in1, wd=str(tmpdir))
+    rout = filesdb.search({}, db=out, wd=str(tmpdir))
+    for r in r1:
+        assert r in rout
+    assert len(rout) == 3
+
+    out = 'outsp.db'
+    filesdb.add(dict(hi=5), wd=str(tmpdir), db=out, filename='5')
+    filesdb.add(dict(rowitem2), db=out, wd=str(tmpdir), copy_mode=True)
+    subprocess.check_call(['filesdb', '--wd={}'.format(tmpdir), '--db={}'.format(out), 'merge', in1])
+    assert subprocess.check_output(['filesdb', '--wd={}'.format(tmpdir), '--db={}'.format(out), 'search']).decode().count('\n') == 4
+
+    out = 'out2.db'
+    filesdb.add(dict(hi=5), wd=str(tmpdir), db=out, filename='5')
+    filesdb.add(dict(rowitem2), db=out, wd=str(tmpdir), copy_mode=True)
+    filesdb.add(dict(hi=100), db=out, wd=str(tmpdir), filename='1')
+    with pytest.raises(RuntimeError):
+        filesdb.merge(in1, out, wd=str(tmpdir))
+
+    out = 'out2sb.db'
+    filesdb.add(dict(hi=5), wd=str(tmpdir), db=out, filename='5')
+    filesdb.add(dict(rowitem2), db=out, wd=str(tmpdir), copy_mode=True)
+    filesdb.add(dict(hi=100), db=out, wd=str(tmpdir), filename='1')
+    with pytest.raises(subprocess.CalledProcessError):
+        subprocess.check_call(['filesdb', '--wd={}'.format(tmpdir), '--db={}'.format(out), 'merge', in1])
